"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;

var _invariant = _interopRequireDefault(require("invariant"));

var _flattenDeep = _interopRequireDefault(require("lodash/flattenDeep"));

var _pick = _interopRequireDefault(require("lodash/pick"));

var _NotAllowed = _interopRequireDefault(require("../errors/NotAllowed"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _objectWithoutProperties(source, excluded) { if (source == null) return {}; var target = _objectWithoutPropertiesLoose(source, excluded); var key, i; if (Object.getOwnPropertySymbols) { var sourceSymbolKeys = Object.getOwnPropertySymbols(source); for (i = 0; i < sourceSymbolKeys.length; i++) { key = sourceSymbolKeys[i]; if (excluded.indexOf(key) >= 0) continue; if (!Object.prototype.propertyIsEnumerable.call(source, key)) continue; target[key] = source[key]; } } return target; }

function _objectWithoutPropertiesLoose(source, excluded) { if (source == null) return {}; var target = {}; var sourceKeys = Object.keys(source); var key, i; for (i = 0; i < sourceKeys.length; i++) { key = sourceKeys[i]; if (excluded.indexOf(key) >= 0) continue; target[key] = source[key]; } return target; }

function ownKeys(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); keys.push.apply(keys, symbols); } return keys; }

function _objectSpread(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys(source, true).forEach(function (key) { _defineProperty(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys(source).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

class PermissionsManager {
  static generateCRUD(resource) {
    return ['create', 'read', 'update', 'delete'].map(action => ({
      permission: `${resource}.${action}`,
      description: `"${action}" action for the "${resource}" resource`
    }));
  }

  constructor(aliases = {
    view: 'read',
    access: 'read'
  }) {
    this.permissions = {};
    this.aliases = aliases || {};
  }

  createAlias(forPermission, toPermission) {
    this.aliases[forPermission] = toPermission;
  }

  getAll() {
    return Object.keys(this.permissions).reduce((acc, permission) => [...acc, _objectSpread({
      permission
    }, (0, _pick.default)(this.permissions[permission], ['description', 'dependencies']))], []);
  }

  has(name) {
    return Object.keys(this.permissions).includes(name);
  }

  define(permission, check, rawConfig = {}) {
    if (Array.isArray(permission)) {
      return permission.forEach((_ref) => {
        let {
          permission: permissionName
        } = _ref,
            itemConfig = _objectWithoutProperties(_ref, ["permission"]);

        return this.define(permissionName, check, _objectSpread({}, rawConfig, {}, itemConfig));
      });
    }

    (0, _invariant.default)(permission && typeof permission === 'string', 'Permissions name is required!');
    (0, _invariant.default)(typeof check === 'function', `Missing required check function for "${permission}" permission!`);

    const config = _objectSpread({
      description: `${permission} permission definition`,
      dependencies: [],
      isValidObject: () => true
    }, rawConfig, {
      check
    });

    if (this.has(permission)) {
      throw new Error(`Can't add permission with a name of ${permission}. Already exists!`);
    }

    const dependencies = (0, _flattenDeep.default)(config.dependencies);
    dependencies.forEach(dependency => {
      if (!this.has(dependency)) {
        throw new Error(`Unable to find permission named "${dependency}" required as a dependency of "${permission}".`);
      }
    });
    this.permissions[permission] = config;
    return this;
  }

  async can(subject, permissionName, object) {
    try {
      await this.check(subject, permissionName, object);
    } catch (err) {
      if (err instanceof _NotAllowed.default) {
        return false;
      }

      throw err;
    }

    return true;
  }

  async checkAny(subject, permissions, object) {
    const result = await Promise.all(permissions.map(permission => this.can(subject, permission, object)));
    const isAllowed = result.some(item => !!item);

    if (!isAllowed) {
      throw new _NotAllowed.default(`Failed permission check.`);
    }
  }

  async checkAll(subject, permissions, object, parallel = true) {
    const walkPermissions = permissions.map(item => {
      if (Array.isArray(item)) {
        return () => this.checkAll(subject, item, object, !parallel);
      }

      return () => this.check(subject, item, object);
    });
    return parallel ? Promise.all(walkPermissions.map(task => task())) : walkPermissions.reduce((promise, task) => promise.then(task), Promise.resolve());
  }

  async check(subject, permissionName, object) {
    (0, _invariant.default)(subject, 'Subject is required!');
    (0, _invariant.default)(permissionName && typeof permissionName === 'string', 'Permission is required!');
    const name = Object.keys(this.aliases).includes(permissionName) ? this.aliases[permissionName] : permissionName;

    if (!this.has(name)) {
      throw new Error(`Unable to find permission named "${name}" when checking access.`);
    }

    const permission = this.permissions[name];

    if (!permission.isValidObject(object)) {
      throw new _NotAllowed.default(`Failed permission check "${name}"! Reason: received invalid object (${object}).`);
    }

    if (permission.dependencies.length) {
      await this.checkAll(subject, permission.dependencies, object);
    }

    const isAllowed = await permission.check(subject, object);

    if (!isAllowed) {
      throw new _NotAllowed.default(`Failed permission check "${name}"! Reason: failed check function.`);
    }
  }

}

var _default = PermissionsManager;
exports.default = _default;